;-------------------------------------------------------------------------------
; * Timer System - Version 4c
; Enhanced MilliSecs wrapper
;-------------------------------------------------------------------------------
; Created by Teasy <TheChange@yahoo.com>
; Last modified on 27 Aug 2011
;-------------------------------------------------------------------------------
; Use this system to easily keep track of time,
; without the need to access MilliSecs directly,
; with the ability to adjust and pause time.
;-------------------------------------------------------------------------------
; Note:
;   These routines require the use of an external global variable called
;   MilliSecs which has to be updated regularly, for example each frame.
;   See the demos section for more information and example usage.
;   Remark the line below if you wish to handle this differently,
;   or perhaps in a different location of your code.

Global MilliSecs

;-------------------------------------------------------------------------------
; Algorithm:
;   This system uses time-hopping which is very accurate
;   but also very math-intensive.
;   Very similar to buffered, unsychronized transfers,
;   time-hopping makes a variable leap nearest to the current time
;   after a timer's interval / duration.
;   This without having to regularly update/poll,
;   plus knowing the number of times a timer's time has elapsed.
;-------------------------------------------------------------------------------
; Usage:
;   handle = NewTimer( interval [, paused ] )
;       Create a timer with specified interval (duration)
;       either paused (true) or running (false, default)
;   DeleteTimer handle
;       Delete timer, free from memory
;
;   num_updates = TimerDone( handle )
;       Update the timer and return the number of times it has elapsed
;       - can be called at any time, but affects other functions too (!)
;   ResetTimer handle [, interval [, paused ] ]
;       Restart timer from current point in time (interval=0, default)
;       - optionally change interval (interval<>0)
;       - optionally change pause state (default=false)
;   SetTimer handle , interval
;       Change timer's interval without altering current point in time
;       - for example to speed up or slow down, or change the time-frame
;   ElapseTimer handle
;       Instantly force timer to elapse at the current point in time
;       - this would for example also trigger a TimerDone()
;   ShiftTimer handle , shift
;       Independently and relatively change a timer's current point in time.
;       - where 1000 means advancing the timer by 1 second (less time)
;         and -5000 means giving the timer 5 seconds extra/more time
;
;   StopTimer handle 
;       Disable an active timer (cancels a pause too)
;   StartTimer handle [, paused ]
;       Activate a stopped timer and run (or optionally pause it)
;
;   PauseTimer handle
;       Pause, remember time (if timer is active)
;   ResumeTimer handle
;       Unpause, restore time (if timer is active)
;
;   interval = TimerInterval( handle )
;       Return current timer's interval (supplied if needed at all)
;   elapsed = TimerElapsed( handle )
;       Return number of milliseconds since timer last elapsed
;       - supports pause/stop, can return values larger than the timer interval
;   remain = TimerRemain( handle )
;       Return number of milliseconds up to when timer will elapse again
;       - supports pause/stop, returns negative values when timer elapsed
;   delta = TimerDelta#( handle )
;       Return a floating point value equal to the interval delta.
;       E.g. 0.0 when the timer was just started
;       and 1.0 when the timer has elapsed, and i.e. 2.0 when elapsed twice.
;       - supports stop and pause (delta stays the same).
;       - the return value can only go over the 1.0 boundary
;         if the timer was not updated (i.e. with TimerDone) directly before.
;
;   state = TimerState( handle )
;       Return -1 if timer is inactive (no matter if paused)
;       Return  0 if timer is paused
;       Return  1 if timer is running
;   paused = TimerPaused( handle )
;       Return True if timer is paused (and active), otherwise False
;   running = TimerRunning( handle )
;       Reverse of TimerPaused (for convenience)
;   stopped = TimerStopped( handle )
;       Return True if timer is not active, otherwise False
;   active = TimerActive( handle )
;       The reverse of TimerStopped (for convenience)
;-------------------------------------------------------------------------------
; Notes:
; - All intervals/durations used (time units and such) are in milliseconds.
; - Negative intervals are supported, reversing all time scales.
; - A time interval of 0 is supported, and gives interesting results,
;   so perhaps there is a use for it :P
;-------------------------------------------------------------------------------
; Known limitations:
;   If you are using Blitz2D/Blitz+/Blitz3D, all variables used to track time
;   are signed 32-bit integers, which means that you can pause a timer
;   up to less than ~25 days (2,147,483,647 ms).
;   As well as any time differences over less than 25 days
;   will wrap around and start from -25 days.
;       Example:
;           a = 2147483647 + 1  ; max + 1 ms
;           b = 0
;           Print a-b
;           WaitKey 
;           End 
;   It is possible to convert all integers to floats,
;   but the timing accuracy will then be compromised.
;   For more help or information on using MilliSecs()
;   see my elaborate MilliSecs tutorial
;   or part 7 of my Blitz Full Foundation tutorial.
;-------------------------------------------------------------------------------
; Example programs:
;   Unremark and modify the following line to see the system in action.
;   Change the number to show another example while looking at the code.

;TimerSystem_Example5


; Example 1: How the system works
Function TimerSystem_Example1()
    Local timer, updates
    MilliSecs = MilliSecs()
    timer = NewTimer( 1000 )
    Print "Timer is set to 1 second (1000 ms)"
    Print "Hit <Spacebar> to check timer"
    Repeat
        MilliSecs = MilliSecs()
        If KeyHit( 57 )  ; Space
            updates = TimerDone( timer )
            Write "Number of timer updates"
            Print " since last spacebar: "+updates
        End If
        Delay 10
    Until KeyHit( 1 )  ; Esc
    End
End Function 


; Example 2: Blinking with FOR
Function TimerSystem_Example2()
    Local timer, updates, update, visible
    MilliSecs = MilliSecs()
    timer = NewTimer( 25 )  ; Interval = 25 ms
    SetBuffer BackBuffer()
    Repeat
        MilliSecs = MilliSecs()
        updates = TimerDone( timer )
        For update = 1 To updates
            visible = Not visible
        Next
        If visible
            Color 255,255,255
        Else
            Color 0,0,0
        EndIf
        Text 0,0 , "[..BLINK..]"
        Flip
        Cls
    Until KeyHit( 1 )  ; Esc
    End
End Function


; Example 3: Blinking with MOD
Function TimerSystem_Example3()
    Local timer, updates, visible
    MilliSecs = MilliSecs()
    timer = NewTimer( 1 )  ; Interval = 1 ms !
    SetBuffer BackBuffer()
    Repeat
        MilliSecs = MilliSecs()
        updates = TimerDone( timer )
        visible = ( visible + updates ) Mod 2
        If visible
            Color 255,255,255
        Else
            Color 0,0,0
        EndIf
        Text 0,0 , "[..BLINK..]"
        Flip False
        Cls
    Until KeyHit( 1 )  ; Esc
    End
End Function 


; Example 4: Moving objects
Function TimerSystem_Example4()
    Local timer_object1, timer_object2
    Local pos_object1, pos_object2
    MilliSecs = MilliSecs()
    timer_object1 = NewTimer( 10 )  ; 1 pixel per 10 ms
    timer_object2 = NewTimer( 15 )  ; 1 pixel per 15 ms
    pos_object1 = 0
    pos_object2 = GraphicsWidth()/2
    SetBuffer BackBuffer()
    Repeat
        ; update time
        MilliSecs = MilliSecs()
        ; update input
        If KeyHit( 57 )  ; space
            If TimerPaused( timer_object1 )
                ResumeTimer timer_object1
                ResumeTimer timer_object2
            Else 
                PauseTimer timer_object1
                PauseTimer timer_object2
            EndIf 
        EndIf 
        ; move objects
        pos_object1 = pos_object1 + TimerDone( timer_object1 )
        pos_object2 = pos_object2 + TimerDone( timer_object2 )
        ; limit objects
        If pos_object1 >= GraphicsWidth()
            pos_object1 = pos_object1 - GraphicsWidth()
        EndIf 
        If pos_object2 >= GraphicsWidth()
            pos_object2 = pos_object2 - GraphicsWidth()
        EndIf 
        ; draw objects
        Color 255,128,128
        Rect pos_object1-50,GraphicsHeight()*1/3-50 , 100,100 , False 
        Color 128,128,255
        Rect pos_object2-50,GraphicsHeight()*2/3-50 , 100,100 , False 
        Color 255,255,255
        Text 0,0 , "Press <Space> to pause/resume"
        Flip
        Cls
        Delay 1
    Until KeyHit( 1 )  ; Esc
    End
End Function 


; Example 5: Demonstrate ALL unique timer functions (!)
Function TimerSystem_Example5()
    Graphics 640,480 , 0,2
    Graphics 300,200 , 0,3
    SetBuffer BackBuffer()
    ClsColor 0,64,128
    Color 255,255,255
    
    MilliSecs = MilliSecs()
    Local min_interval = 1
    Local max_interval = 8192
    Local keyrepeat_timer = NewTimer( 150 )  ; left/right key hold down slowdown
    Local timer = NewTimer( 4096 )  ; new timer test
    Local completed  ; times completed (elapsed)
    Local check = True  ; check for timerdone
    
    Local interval
    Repeat
        MilliSecs = MilliSecs()
        
        If KeyHit(57)  ; space
            If TimerPaused( timer )
                ResumeTimer timer
            Else
                PauseTimer timer
            End If
        End If
        If KeyHit(14)  ; backspace
            If TimerStopped( timer )
                StartTimer timer
            Else
                StopTimer timer
            End If
        End If
        If KeyHit(28)  ; enter
            check = Not check
        End If
        
        If KeyHit(15)  ; tab
            SetTimer timer , -TimerInterval( timer )
        End If
        If KeyHit(42) Or KeyHit(54)  ; left/right shift
            ElapseTimer timer
        End If
        If KeyHit(200)  ; up
            ShiftTimer timer , TimerInterval( timer ) / 4
        End If
        If KeyHit(208)  ; down
            ShiftTimer timer , -TimerInterval( timer ) / 4
        End If
        
        Local keyrepeat = TimerDone( keyrepeat_timer )
        If KeyDown(203) And keyrepeat Or KeyHit(203)  ; left
            ResetTimer keyrepeat_timer
            interval = TimerInterval( timer )
            If interval > min_interval
                SetTimer timer , interval / 2
            ElseIf interval > 0
            ElseIf interval > -max_interval
                SetTimer timer , interval * 2
            End If
        End If
        If KeyDown(205) And keyrepeat Or KeyHit(205)  ; right
            ResetTimer keyrepeat_timer
            interval = TimerInterval( timer )
            If interval < -min_interval
                SetTimer timer , interval / 2
            ElseIf interval < 0
            ElseIf interval < max_interval
                SetTimer timer , interval * 2
            End If
        End If
        
        Text 0,0 , "Timer Remain: "+TimerRemain( timer )
        Text 0,10 , "Timer Elapsed: "+TimerElapsed( timer )+"   <shift>"
        Text 0,20 , "Timer Paused: "+TimerPaused( timer )+"   <space>"
        Text 0,30 , "Timer Stopped: "+TimerStopped( timer )+"   <backspace>"
        Text 0,40 , "Timer Interval: "+TimerInterval( timer )+"   <left/right>"
        Text 0,50 , "negate interval   <tab>"
        Local delta# = TimerDelta( timer )
        Text 0,60 , "Shift Timer fwd/bwd   <up/down>"
        Text 0,70 , "Timer Delta: "+delta
        Local deltachar$ = Mid( "<>" , (delta>0)+1 , 1 )
        Text 0,80 , String( deltachar , Abs( delta ) * 10 )
        Text 0,90 , "|" + String( ".",8 ) + "| = "+Abs( TimerInterval( timer ) )+"ms"
        If check
            completed = completed + TimerDone( timer )
            Text 0,110 , "update timer: Yes   <enter>"
        Else
            Text 0,110 , "update timer: No   <enter>"
        End If
        Text 0,120 , "times completed: "+completed
        
        Text 0,140 , "quit   <esc>"
        Flip
        Cls
        Delay 5
    Until KeyHit(1)  ; esc
    DeleteTimer timer
    DeleteTimer keyrepeat_timer
    End
End Function

;-------------------------------------------------------------------------------
; = TimerSystem Code =
;-------------------------------------------------------------------------------
; (Internal) Low level (No error checking)
Type Timer_Object_
    Field LastUpdate
    Field Duration
    Field Paused,Active
    Field Difference
End Type
Global Timer_Object__Count
Function Create_Timer_Object.Timer_Object_( Started , Duration , Paused )
    Local p.Timer_Object_ = New Timer_Object_
    p\LastUpdate = Started
    p\Duration = Duration
    p\Paused = Paused
    p\Active = True
    p\Difference = 0
    Timer_Object__Count = Timer_Object__Count + 1
    Return p
End Function
Function Delete_Timer_Object( p.Timer_Object_ )
    Delete p
    Timer_Object__Count = Timer_Object__Count - 1
End Function
;-------------------------------------------------------------------------------
; (External) Interface (No error checking)
Function NewTimer( TimeOut , Paused=False )
    Return Handle Create_Timer_Object( MilliSecs , TimeOut , Paused )
End Function
Function DeleteTimer( Timer_Handle )
    Delete_Timer_Object Object.Timer_Object_( Timer_Handle )
End Function
Function TimerDone( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Timer must not be paused and must be running
    If p\Active = True And p\Paused = False
        ; How much time has passed
        Local Difference = MilliSecs - p\LastUpdate
        ; At least one update must have passed
        If Difference >= p\Duration
            ; How many updates in time passed
            Local Updates = Floor( Float Difference / p\Duration )
            ; Set time ahead number of updates done
            p\LastUpdate = MilliSecs - ( Difference - Updates * p\Duration )
            ; Return number of times timer's time has elapsed
            Return Updates
        End If
    End If
End Function
Function ResetTimer( Timer_Handle , TimeOut=0 , Paused=False )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Sync time to now
    p\LastUpdate = MilliSecs
    ; If change in interval/duration specified
    If TimeOut <> 0
        ; Change interval/duration
        p\Duration = TimeOut
    EndIf
    ; Change pause state (and forget time difference after pause)
    p\Paused = Paused
    p\Difference = 0
End Function
Function SetTimer( Timer_Handle , TimeOut )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Change interval/duration
    p\Duration = TimeOut
End Function
Function ElapseTimer( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    If p\Duration < 0
        ; allow a timer with negative interval to elapse in the opposite dir
        p\LastUpdate = MilliSecs
    Else
        ; Force elapse of timer at current point in time
        p\LastUpdate = MilliSecs - p\Duration
    End If
End Function
Function ShiftTimer( Timer_Handle , Time_Shift )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    If p\Active
        p\LastUpdate = p\LastUpdate - Time_Shift
    End If
End Function
Function PauseTimer( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Can't pause a stopped timer (must be active)
    If p\Active
        p\Paused = True
        ; Remember difference between LastUpdate and MilliSecs (for unpause)
        p\Difference = MilliSecs - p\LastUpdate
    EndIf
End Function
Function ResumeTimer( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Can't unpause a stopped timer (must be active)
    If p\Active
        p\Paused = False
        ; Difference is restored: Timer is set 'Difference' behind MilliSecs
        p\LastUpdate = MilliSecs - p\Difference
    EndIf
End Function
Function StopTimer( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    If p\Active
        If p\Paused
            ResumeTimer Timer_Handle
        End If
        ; Deactivate timer
        p\Active = False
    End If
End Function
Function StartTimer( Timer_Handle , Paused=False )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    If Not p\Active
        ; Activate timer
        p\Active = True
        ResetTimer Timer_Handle , 0 , Paused
    End If
End Function
Function TimerInterval( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    Return p\Duration
End Function 
Function TimerElapsed( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    If p\Active
        If p\Paused
            Return p\Difference
        Else
            Return MilliSecs - p\LastUpdate
        End If
    Else
        Return 0
    End If
End Function
Function TimerRemain( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    If p\Active
        If p\Paused
            Return p\Duration - p\Difference
        Else
            Return p\LastUpdate + p\Duration - MilliSecs
        End If
    Else
        Return p\Duration
    End If
End Function
Function TimerDelta#( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    If p\Active
        If p\Paused
            Return Float p\Difference / p\Duration
        Else
            Return Float ( MilliSecs - p\LastUpdate ) / p\Duration
        End If
    Else
        Return 0
    End If
End Function
Function TimerState( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Timer is inactive
    If Not p\Active
        Return -1
    ; Timer is paused
    ElseIf p\Paused
        Return 0
    ; Timer is running
    Else 
        Return 1
    EndIf 
End Function
Function TimerActive( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    Return p\Active
End Function
Function TimerStopped( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    Return Not p\Active
End Function
Function TimerRunning( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    If p\Active
        Return Not p\Paused
    Else
        Return False
    End If
End Function
Function TimerPaused( Timer_Handle )
    Local p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    If p\Active
        Return p\Paused
    Else
        Return False
    End If
End Function

;-------------------------------------------------------------------------------
; Updates
;-------------------------------------------------------------------------------
; 05 Nov 2006
; - added ability to combine with Blitz' default timing system
; 03 May 2007
; - added settimer that changes duration/interval without resetting time memory
; 03 Jul 2007
; - added timerinterval to easily retrieve interval if needed
; 10 Aug 2007
; - updated documentation & demos
; - renamed Duration to Interval (for clarity)
; - added TimerState, TimerStopped and TimerPaused
; 11 Nov 2007
; - added TimerElapsed to return number of ms since last elapsed
; - added TimerDelta to 'percentage' (0..1) of interval elapsed since last elapsed
; 11 Apr 2008
; - added Local variables to prevent/fix mismatch/mixing of Global influences
; 09 Sep 2008
; - added Float fix in TimerDone to prevent division by zero
; 01 Feb 2009
; - added TimerRemain which returns number of millisecs until next elapse time
; 23 Dec 2010
; - added TimerDelta command to documentation
; 27 Aug 2011
;   (these update may affect existing code, so i've updated the version number)
; - added pause-awareness to TimerElapsed, TimerRemain, TimerDelta
; - added stop-awareness to all functions (!)
; - added TimerRunning (not paused) and TimerActive (not stopped)
; - added paused parameter to StartTimer
; - fixed a rounding error in TimerDone when a timer elapsed multiple times
; - added support for negative intervals (don't know if it's nice in practice)
; - fixed ElapseTimer to work with a negative timer
; - added ShiftTimer to advance the timer (less time) or to give it more time
; - updated all of the help text (i think)
; - added a 5th demo to test the stability of these updates
;   and to demonstrate all of the functions (!)
;-------------------------------------------------------------------------------

;~IDEal Editor Parameters:
;~C#Blitz3D198